//	VirtualDub - Video processing and capture application
//	Copyright (C) 1998-2001 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include <string.h>
#include <windows.h>
#include "oshelper.h"
#include "helpfile.h"
#include "gui.h"
#include "resource.h"

extern const char g_szCapture[];
static const char g_szCaptureWarn[]="Disabled Warnings";

#define CWF_PINNACLE		(0x00000001L)
#define CWF_ZORAN			(0x00000002L)
#define CWF_BROOKTREE		(0x00000004L)

static long g_capwarnFlags = -1;

static void CaptureWarnInit() {
	if (g_capwarnFlags < 0)
		if (!QueryConfigDword(g_szCapture, g_szCaptureWarn, (DWORD *)&g_capwarnFlags))
			g_capwarnFlags = 0;
}

static void CaptureWarnDisable(long f) {
	g_capwarnFlags |= f;
	SetConfigDword(g_szCapture, g_szCaptureWarn, g_capwarnFlags);
}

void CaptureWarnCheckDriver(HWND hwnd, const char *s) {
	CaptureWarnInit();

	if (!(g_capwarnFlags & CWF_PINNACLE) && (strstr(s, "Pinnacle") || strstr(s, "miroVIDEO"))) {

		if (IDYES == guiMessageBox(hwnd, IDS_WARN_CAPTURE_PINNACLE, IDS_WARN_CAPTURE_PINNACLE_CAP, MB_YESNO)) {
			HelpPopup(hwnd, IDH_CAPWARN_PINNACLE);
		}

		CaptureWarnDisable(CWF_PINNACLE);
	}
}

void CaptureWarnCheckDrivers(HWND hwnd) {
	HANDLE hFind;
	WIN32_FIND_DATA wfd;
	char szPath[MAX_PATH], *s;

	if (!(g_capwarnFlags & CWF_ZORAN)) {
		GetWindowsDirectory(szPath, sizeof szPath);
		s = szPath;
		while(*s) ++s;
		if (s[-1] != '\\')
			*s++ = '\\';

		strcpy(s, "system\\h20capt.dll");

		hFind = FindFirstFile(szPath, &wfd);

		if (hFind == INVALID_HANDLE_VALUE) {
			strcpy(s, "system\\h22capt.dll");

			hFind = FindFirstFile(szPath, &wfd);
		}

		if (hFind != INVALID_HANDLE_VALUE) {
			FindClose(hFind);
			
			CaptureWarnInit();

			if (IDYES == guiMessageBox(hwnd, IDS_WARN_CAPTURE_ZORAN, IDS_WARN_CAPTURE_ZORAN_CAP, MB_YESNO)) {
				HelpPopup(hwnd, IDH_CAPWARN_ZORAN);
			}
			CaptureWarnDisable(CWF_ZORAN);
		}
	}

	if (!(g_capwarnFlags & CWF_BROOKTREE)) {
		GetWindowsDirectory(szPath, sizeof szPath);
		s = szPath;
		while(*s) ++s;
		if (s[-1] != '\\')
			*s++ = '\\';

		strcpy(s, "system\\bt848_32.dll");

		hFind = FindFirstFile(szPath, &wfd);

		if (hFind == INVALID_HANDLE_VALUE) {
			strcpy(s, "system32\\bt848_32.dll");

			hFind = FindFirstFile(szPath, &wfd);
		}

		if (hFind != INVALID_HANDLE_VALUE) {
			FindClose(hFind);
			
			CaptureWarnInit();

			if (IDYES == guiMessageBox(hwnd, IDS_WARN_CAPTURE_BT848, IDS_WARN_CAPTURE_BT848_CAP, MB_YESNO)) {
				HelpPopup(hwnd, IDH_CAPWARN_BROOKTREE);
			}
			CaptureWarnDisable(CWF_BROOKTREE);
		}
	}
}

